import time
from .service_access_point import DENRequest, PriorityLevel
from ...facilities.decentralized_environmental_notification_service.den_service import (
    DecentralizedEnvironmentalNotificationService,
)
from ...btp.router import Router as BTPRouter


class EmergencyVehicleApproachingService:
    """
    The Emergency Vehicle Approaching Service is the class that triggers a
    DENM transmission notifying about the approach of an emergency vehicle.
    It's an example of a partial implementation of the Road Hazard Signalling (RHS)
    Application.
    It expects to receive the position of the approaching vehicle and its characterstics
    to trigger the periodic and temporary transmission of DENM messages.

    Attributes
    ----------
    den_service : DecentralizedEnvironmentalNotificationService
        Reference to the DEN Basic Service for handling DENM transmissions.
    denm_interval : int
        Time interval (in milliseconds) between two consecutive DENM messages.
    denm_duration : int
        Total duration (in milliseconds) for which DENM messages will be transmitted.
    priority_level : int
        Priority level of the DENM messages being sent.
    detection_time : int
        Timestamp of the hazard detection event in milliseconds since 2004-01-01T00:00:00Z.
    event_position : dict
        Dictionary containing the geographic position of the hazard, including latitude, longitude,
        altitude, and confidence values.
    """

    def __init__(
        self, btp_router: BTPRouter, vehicle_data, duration: int = 10000
    ) -> None:
        """
        Initialize the Emergency Vehicle Approaching Service.

        Parameters
        ----------
         Parameters
        ----------
        btp_router : BTPRouter
            Reference to the BTP Router for communication.
        vehicle_data : Any
            Vehicle-specific data used for initializing the DEN service.
        duration : int, optional
            Duration in milliseconds for which DENM messages will be transmitted
            (default is 10 seconds).
        """
        self.den_service = (
            DecentralizedEnvironmentalNotificationService(  # passar per referencia
                btp_router, vehicle_data
            )
        )
        self.denm_duration = duration  # [ms] 10 seconds
        self.denm_interval = 100
        self.priority_level = PriorityLevel.WARNING
        # Get DENM data to simulate the hazard detection
        self.detection_time = int((time.time() - 1072911600) * 1000)
        self.event_position = {
            "latitude": 900000001,
            "longitude": 1800000001,
            "positionConfidenceEllipse": {
                "semiMajorConfidence": 4095,
                "semiMinorConfidence": 4095,
                "semiMajorOrientation": 3601,
            },
            "altitude": {"altitudeValue": 800001, "altitudeConfidence": "unavailable"},
        }

    def trigger_denm_sending(self, tpv: dict) -> None:
        """
        Trigger the DENM sending process using the facilities layer.

        Parameters
        ----------
        tpv : dict
            Dictionary containing the current position and altitude information.
            Expected keys include:
                - "lat" : Latitude in decimal degrees.
                - "lon" : Longitude in decimal degrees.
                - "altHAE" : Altitude in meters above the WGS-84 ellipsoid.
        """
        request = DENRequest()
        if "lat" in tpv.keys():
            self.event_position["latitude"] = int(tpv["lat"] * 10000000)
        if "lon" in tpv.keys():
            self.event_position["longitude"] = int(tpv["lon"] * 10000000)
        if "altHAE" in tpv.keys():
            alt = int(tpv["altHAE"] * 100)
            if alt < -800000:
                self.event_position["altitude"]["altitudeValue"] = -100000
            elif alt > 613000:
                self.event_position["altitude"]["altitudeValue"] = 800000
            else:
                self.event_position["altitude"]["altitudeValue"] = int(
                    tpv["altHAE"] * 100
                )

        request.with_emergency_vehicle_approaching(self)
        self.den_service.denm_transmission_management.request_denm_sending(request)



from __future__ import annotations
import socket
import platform
from collections.abc import Callable
import threading
from .exceptions import (
    WindowsNotSupportedException,
    InvalidMACAddressException,
    PacketTooLongException,
)
from .link_layer import LinkLayer


def raise_exception_if_windows(func):
    """
    Decorator to raise an exception if the function is called on a Windows system.
    """

    def wrapper(*args, **kwargs):
        if platform.system() == "Windows":
            raise WindowsNotSupportedException(
                "Can't currently run RawLinkLayer on a Windows system"
            )
        return func(*args, **kwargs)

    return wrapper


@raise_exception_if_windows
class RawLinkLayer(LinkLayer):
    """
    Link Layer class to connect layer 2 and layer 3.

    Attributes
    ----------
    receive_callback : Callable[[bytes], None]
        Callback function to receive packets.
    sock : socket
        Socket to send and receive packets.
    mac_address : bytes
        MAC address of the interface.

    Methods
    -------
    send(bytes)
        Send a packet to the LL.
    receive()
        Receive a packet from the LL. (To be called in a thread)


    """

    def __init__(
        self, iface: str, mac_address: bytes, receive_callback: Callable[[bytes], None]
    ) -> None:
        """
        Create a Link Layer object.

        Parameters
        ----------
        iface : str
            Interface to use. (Linux only)
        mac_address : bytes
            MAC address of the interface.
        receive_callback : Callable[[bytes], None]
            Callback function to receive packets.

        Raises
        ------
        OSError:
            If the interface is not found.
        Exception:
            If the MAC address is not 6 bytes long.
        """
        super().__init__(receive_callback)
        self.sock = socket.socket(
            socket.AF_PACKET, socket.SOCK_RAW, socket.htons(0x8947)
        )
        self.sock.bind((iface, 0))
        if len(mac_address) != 6:
            raise InvalidMACAddressException("MAC address must be 6 bytes long")
        self.mac_address = mac_address
        self.receiving_thread = threading.Thread(target=self.receive, daemon=True)
        self.receiving_thread.start()

    def __del__(self) -> None:
        """
        Close the socket.
        """
        self.sock.close()

    def send(self, packet: bytes) -> None:
        """
        Send a packet to the LL.

        Parameters
        ----------
        packet : bytes
            Packet to send.
        """
        dest = b"\xff\xff\xff\xff\xff\xff"
        ethertype = b"\x89\x47"
        packet = dest + self.mac_address + ethertype + packet
        if len(packet) > 1500:
            raise PacketTooLongException("Packet too long")
        self.sock.send(packet)

    def receive(self) -> None:
        """
        Receive a packet from the LL.
        """
        while True:
            try:
                m = self.sock.recv(1500)
                try:
                    if m[0:6] == self.mac_address:
                        self.receive_callback(m[14:])
                    elif (
                        m[0:6] == b"\xff\xff\xff\xff\xff\xff"
                        and m[6:12] != self.mac_address
                    ):
                        self.receive_callback(m[14:])
                    # if m[0:6] == b'\xff\xff\xff\xff\xff\xff' and m[6:12] == self.mac_address:
                except NotImplementedError as e:
                    print("Error decoding packet: " + str(e))
            except OSError:
                break



"""
General Constant for LDM
"""
import os
library_folder = os.path.dirname(os.path.abspath(__file__))
file_path = os.path.join(library_folder, "data_containers.json")
LDM_STORAGE_PATH = file_path
REFERENCE_ITS_TIMESTAMP = 1072911600
MAINTENANCE_AREA_MAX_ALTITUDE_DIFFERENCE = 15
EATH_RADIUS = 6371000.0


"""
Consatants for Interface LDM Service. As mentioned in ETSI EN 302 895 V1.1.1 (2014-09).
"""
DATA_OBJECT_FIELD_NAME = "dataObject"

"""
Constants for Interface LDM Maintenance. As mentioned in ETSI EN 302 895 V1.1.1 (2014-09).
"""
NEW_DATA_RECIEVED = 0
NO_NEW_DATA_RECIEVED = 1
MAINTENANCE_AREA_MAX_ALTITUDE_DIFFERENCE = 15

"""
Constants for Interface IF.LDM.2. As mentioned in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.1.1.1
"""

RESULT_SUCCESS = "Successful"
RESULT_INVALID_ITS_AID = "Fail: Invalid ITS-AID"
RESULT_UNABLE_APPLICATION_AUTHENTICATION = "Fail: Unable to authenticate application"
RESULT_NOT_AUTHORIZED = "Fail: Application not authorized for requested permissions"

"""
Constants for Interface IF.LDM.3, As mentioned in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.2
"""
REGISTER_DATA_PROVIDER_RESULT_ACCEPTED = "accepted"
REGISTER_DATA_PROVIDER_RESULT_REJECTED = "rejected"

ADD_DATA_PROVIDER_RESULT_ACCEPTED = "Accepted"
ADD_DATA_PROVIDER_RESULT_REJECTED = "unsuccessful"

UPDATE_DATA_PROVIDER_RESULT_ACCEPTED = "UpdateSuccessful"
UPDATE_DATA_PROVIDER_RESULT_REJECTED_DOES_NOT_EXIST = "UpdateRejectedDoesNotExist"
UPDATE_DATA_PROVIDER_RESULT_REJECTED_INCONSISTENT_TYPE = (
    "UpdateRejectedInconsistentType"
)

DELETE_DATA_PROVIDER_RESULT_ACCEPTED = "DeleteSuccessful"
DELETE_DATA_PROVIDER_RESULT_REJECTED = "DeleteRejected"

"""
Constants for the Interface IF.LDM.4. As mentioned in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3
"""

REGISTER_DATA_CONSUMER_RESULT_ACCEPTED = "Accepted"
REGISTER_DATA_CONSUMER_RESULT_REJECTED = "Rejected"

REQUEST_DATA_OBJECT_ACCEPTED = "Accepted"
REQUEST_DATA_OBJECT_INVALID_ITS_AID = "Fail: Invalid ITS-AID"
REQUEST_DATA_OBJECT_INVALID_DATA_OBJECT_TYPE = "Fail: Invalid Data Object Type"
REQUEST_DATA_OBJECT_INVALID_PRIORITY = "Fail: Invalid Priority"
REQUEST_DATA_OBJECT_INVALID_FILTER = "Fail: Invalid Filter"
REQUEST_DATA_OBJECT_INVALID_ORDER = "Fail: Invalid Order"

"""
ITS-AID defined in standard ETSI TS 102 965 V2.1.1 (2021-11).
TODO: This might be wrong! ITS-AID is standarized in ETSI TS 102 965 V2.1.1 (2021-11);
CA_BASIC_SERVICE = 36
DEN_BASIC_SERVICE = 37
TLM_BASIC_SERVICE = 137
RLT_SERVICE = 138
IVI_SERVICE = 139
GN_MGMT = 141
SA_SERVICE = 540801
CRL_SERVICE = 622
SECURED_CERTIFICATE_REQUEST_SERVICE = 623
CTL_SERVICE = 624
GPS_SERVICE = 625
GPC_SERVICE = 540802
CP_SERVICE = 639
VRU_SERVICE = 638
TLC_REQUEST_SERVICE = 140
IMZ_SERVICE = 640

"""

DENM = 1
CAM = 2
POI = 3
SPATEM = 4
MAPEM = 5
IVIM = 6
EV_RSR = 7
TISTPGTRANSACTION = 8
SREM = 9
SSEM = 10
EVSCN = 11
SAEM = 12
RTCMEM = 13
CPM = 14
IMZM = 15
VAM = 16
DSM = 17
PCIM = 18
PCVM = 19
MCM = 20
PAM = 21

DATA_OBJECT_TYPE_ID = {
    1: "denm",
    2: "cam",
    3: "poi",
    4: "spatem",
    5: "mapem",
    6: "ivim",
    7: "ev-rsr",
    8: "tistpgtransaction",
    9: "srem",
    10: "ssem",
    11: "evcsn",
    12: "saem",
    13: "rtcmem",
    14: "cpm",
    15: "imzm",
    16: "vam",
    17: "dsm",
    18: "pcim",
    19: "pcvm",
    20: "mcm",
    21: "pam",
}

VALID_ITS_AID = {
    DENM,
    CAM,
    POI,
    SPATEM,
    MAPEM,
    IVIM,
    EV_RSR,
    TISTPGTRANSACTION,
    SREM,
    SSEM,
    EVSCN,
    SAEM,
    RTCMEM,
    CPM,
    IMZM,
    VAM,
    DSM,
    PCIM,
    PCVM,
    MCM,
    PAM
}


UNKNOWN = 0
PEDESTRIAN = 1
CYCLIST = 2
MOPED = 3
MOTORCYCLE = 4
PASSENGER_CAR = 5
BUS = 6
LIGHT_TRUCK = 7
HEAVY_TRUCK = 8
TRAILER = 9
SPCIAL_VEHICLES = 10
TRAM = 11
ROAD_SIDE_UNIT = 15

"""
Operator contant
"""
OPERATOR_MAPPING = {
    "==": lambda x, y: x == y,
    "!=": lambda x, y: x != y,
    ">": lambda x, y: x > y,
    "<": lambda x, y: x < y,
    ">=": lambda x, y: x >= y,
    "<=": lambda x, y: x <= y,
    "like": lambda x, y: x in y,
    "notlike": lambda x, y: x not in y
}



class ECIESBackend:
    """
    Interface for ECIES backends.
    """

    def create_key(self) -> int:
        """
        Creates a new key pair. Returns the identifier of the key.
        """
        raise NotImplementedError("All the ECIES backends should implement this")

    def get_public_key(self, identifier: int) -> dict:
        """
        Returns the public key of the key pair with the id.
        It returns as a dict to be oer encoded as PublicEncryptionKey.
        """
        raise NotImplementedError("All the ECIES backends should implement this")

    def encrypt(self, data: bytes, public_key: dict) -> bytes:
        """
        Encrypts the array of bytes. With the public key.
        The public key has to be received as PublicEncryptionKey.
        """
        raise NotImplementedError("All the ECIES backends should implement this")

    def decrypt(self, data: bytes, identifier: int) -> bytes:
        """
        Decrypts the array of bytes. With the private key.
        """
        raise NotImplementedError("All the ECIES backends should implement this")



import time
import logging

from ...facilities.local_dynamic_map.ldm_classes import (
    AddDataProviderReq,
    Location,
    RegisterDataProviderReq,
    TimeValidity,
    TimestampIts,
)
from ...facilities.local_dynamic_map.ldm_constants import VAM

from ...facilities.local_dynamic_map.ldm_facility import LDMFacility


class VRUBasicServiceLDM:
    """
    Class to simplify the operation of the LDM for the VRU Basic Service Component.

    It will handle registry and adding data to the LDM.

    Attributes
    ----------
    ldm_facility: LDMFacility
        Local Dynamic Map Facility
    access_permissions: list
        List containing the application ID of all the applications that want to be accessed.
    time_validity: int
        Time that the messages stored in the LDM will be mantained.
    """

    def __init__(self, ldm: LDMFacility, access_permissions: list, time_validity: int):
        self.logging = logging.getLogger("vru_basic_service")
        self.ldm_if_ldm_3 = ldm.if_ldm_3
        self.access_permissions = access_permissions
        self.time_validity = time_validity
        self.ldm_if_ldm_3.register_data_provider(
            RegisterDataProviderReq(
                application_id=VAM,
                access_permissions=self.access_permissions,
                time_validity=TimeValidity(self.time_validity),
            )
        )

    def add_provider_data_to_ldm(self, vam: dict) -> None:
        """
        Function to add VRU Awareness Messages to the LDM.

        Parameters
        ----------
        vam: dict
            VRU Awareness Message in a python dictionary format

        Returns
        --------
        None
        """
        data = AddDataProviderReq(
            application_id=VAM,
            time_stamp=TimestampIts().insert_unix_timestamp(time.time()),
            location=Location.location_builder_circle(
                latitude=vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["latitude"],
                longitude=vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["longitude"],
                altitude=vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["altitude"][
                    "altitudeValue"
                ],
                radius=0,
            ),
            data_object=vam,
            time_validity=TimeValidity(self.time_validity),
        )
        self.logging.debug(
            "Adding VAM message to LDM with; time_stamp=%d latitude=%f longitude=%f altitude=%d time_validity=%d",
            int(TimestampIts().insert_unix_timestamp(time.time())),
            vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["latitude"],
            vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["longitude"],
            vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["altitude"]["altitudeValue"],
            self.time_validity
        )

        response = self.ldm_if_ldm_3.add_provider_data(data)
        if not isinstance(response.data_object_id, int):
            raise TypeError(response.data_object_id)



from .sn_sap import SNSIGNRequest, SNSIGNConfirm
from .certificate import OwnCertificate
from .security_coder import SecurityCoder
from .ecdsa_backend import ECDSABackend
from ..utils.time_service import TimeService


class CooperativeAwarenessMessageSecurityHandler:
    """
    Class that handles the signing of CAMs according the standard ETSI TS 103 097 V2.1.1 (2021-10) 7.1.1

    Attributes
    ----------

    """

    def __init__(self, coder: SecurityCoder, ecdsa_backend: ECDSABackend) -> None:
        self.coder: SecurityCoder = coder
        # self.ecdsa_backend : ECDSABackend = ecdsa_backend
        self.last_signer_full_certificate_time: float = 0
        self.requested_own_certificate: bool = False

    def sign(self, signed_data: dict, certificate: OwnCertificate) -> bytes:
        """
        Sign the CAM.
        """
        tobesigned: bytes = self.coder.encode_to_be_signed_data(
            signed_data["content"][1]["tbsData"]
        )
        # at_item : OwnCertificate = self.get_present_at_for_signging(request.its_aid)

        signed_data["content"][1]["signer"][1] = certificate.as_hashedid8()
        signed_data["content"][1]["signature"] = certificate.sign_message(tobesigned)

    def set_up_signer(self, certificate: OwnCertificate) -> None:
        """
        Set up the signer. According to the standard rules:
        - As default, the choice digest shall be included.
        - The choice certificate shall be included once, one second after the last inclusion of the choice
        certificate.
        - If the ITS-S receives a CAM signed by a previously unknown AT, it shall include the choice
        certificate immediately in its next CAM, instead of including the choice digest. In this case, the
        timer for the next inclusion of the choice certificate shall be restarted.
        - If an ITS-S receives a CAM that includes a tbsdata.headerInfo component of type
        inlineP2pcdRequest, then the ITS-S shall evaluate the list of certificate digests included in that
        component: If the ITS-S finds a certificate digest of the currently used authorization ticket in that list, it
        shall include the choice certificate immediately in its next CAM, instead of including the choice
        digest.
        """
        signer: tuple = ("digest", certificate.as_hashedid8())
        current_time = TimeService.time()
        if (
            current_time - self.last_signer_full_certificate_time > 1
            or self.requested_own_certificate
        ):
            self.last_signer_full_certificate_time = current_time
            signer = ("certificate", certificate.encode())
        return signer


class SignService:
    """
    Sign service class to sign and verify messages. Follows the specification of ETSI TS 102 723-8 V1.1.1 (2016-04) standard.

    Attributes
    ----------
    ecdsa_backend : ECDSABackend
        ECDSA backend to use.
    unknown_ats : List[bytes]
        List of unknown ATs. Each AT is represented by its certificate hashedId3.
    knwon_ats : dict[bytes, Certificate]
        Dictionary of known ATs, indexed by the AT's certificate hashId8. Each AT is represented by a dictionary with the following keys: 'verifying_key', 'verified', 'certificate', 'backend_id'.
    requested_ats : List[bytes]
        List of requested ATs. Each AT is represented by its certificate hashedId3.
    known_aas : Dict[bytes, Certificate]
        Dictionary of AA. Indexed by the AA HashedId8. Each AA is represented by a Certificate object.
    root_ca : dict[bytes, Certificate]
        Dictionary of Root CA. Indexed by the Root CA HashedId8. Each Root CA is represented by a dictionary with the following keys: 'verifying_key', 'verified', 'certificate'.
    present_ats : dict[int, OwnCertificate]
        The present ats stores the ATs to use for signing. Indexed by the AT's backend id. Each AT is stored as a Certificate.
    """

    def __init__(self, backend: ECDSABackend, security_coder: SecurityCoder) -> None:
        """
        Initialize the Sign Service.
        """
        self.ecdsa_backend: ECDSABackend = backend
        self.coder: SecurityCoder = security_coder
        self.knwon_ats = {}
        self.unknown_ats = {}
        self.requested_ats = {}
        self.knwon_aas = {}
        self.root_ca = {}
        self.present_ats = {}

    def sign_request(self, request: SNSIGNRequest) -> SNSIGNConfirm:
        """
        Sign a SNSIGNRequest.
        """
        if request.its_aid == 36:
            raise NotImplementedError("CA signing is not implemented")
        elif request.its_aid == 37:
            raise NotImplementedError("DEN signing is not implemented")
        elif request.its_aid == 137:
            raise NotImplementedError("TLM signing is not implemented (SPATEM)")
        elif request.its_aid == 138:
            raise NotImplementedError("RLT signing is not implemented (MAPEM)")
        elif request.its_aid == 139:
            raise NotImplementedError("IVI signing is not implemented (IVIM)")
        elif request.its_aid == 141:
            raise NotImplementedError(
                "GeoNetworking Management Communications (GN-MGMT)  signing is not implemented (SPATEM)"
            )
        elif request.its_aid == 540 or request.its_aid == 801:
            raise NotImplementedError("SA service signing is not implemented")
        elif request.its_aid == 639:
            raise NotImplementedError("CP signing is not implemented")
        elif request.its_aid == 638:
            raise NotImplementedError("VRU signing is not implemented")
        else:
            raise NotImplementedError(
                "Security profile for the specified message not implemented"
            )

    def sign_cam(self, request: SNSIGNRequest) -> SNSIGNConfirm:
        """
        Sign a CAM according the standard ETSI TS 103 097 V2.1.1 (2021-10) 7.1.1
        """
        sigend_data_dict = {
            "protocolVersion": 3,
            "content": (
                "signedData",
                {
                    "hashId": "sha256",
                    "tbsData": {
                        "payload": {
                            "data": {
                                "protocolVersion": 3,
                                "content": ("unsecuredData", request.tbs_message),
                            }
                        },
                        "headerInfo": {
                            "psid": request.its_aid,
                            # "generationTime": 0,
                            # "expireTime": 0
                        },
                    },
                    "signer": ("digest", b"\x00\x00\x00\x00\x00\x00\x00\x00"),
                    "signature": (
                        "ecdsaNistP256Signature",
                        {
                            "rSig": ("fill", None),
                            "sSig": (0xA495991B7852B855).to_bytes(32, byteorder="big"),
                        },
                    ),
                },
            ),
        }
        if len(self.unknown_ats) > 0:
            sigend_data_dict["content"][1]["tbsData"]["headerInfo"][
                "inlineP2pcdRequest"
            ] = self.unknown_ats
        if len(self.requested_ats) > 0:
            sigend_data_dict["content"][1]["tbsData"]["headerInfo"][
                "requestedCertificate"
            ] = self.get_known_at_for_request(self.requested_ats.pop(0))

        tobesigned: bytes = self.coder.encode_to_be_signed_data(
            sigend_data_dict["content"][1]["tbsData"]
        )
        at_item: OwnCertificate = self.get_present_at_for_signging(request.its_aid)
        sigend_data_dict["content"][1]["signer"] = ("digest", at_item.as_hashedid8())
        sigend_data_dict["content"][1]["signature"] = at_item.sign_message(tobesigned)

        sec_message = self.coder.encode_etsi_ts_103097_data_signed(sigend_data_dict)
        confirm = SNSIGNConfirm(
            sec_message=sec_message, sec_message_length=len(sec_message)
        )

        return confirm

    def get_present_at_for_signging(self, its_aid: int) -> OwnCertificate:
        """
        Get the present AT for a given ITS-AID.
        """
        for cert in self.present_ats.items():
            if its_aid in cert[1].get_list_of_its_aid():
                return cert[1]
        return None

    def get_known_at_for_request(self, hashedid3: bytes) -> dict:
        """
        Get the known AT for a given hashedId3.
        """
        raise NotImplementedError(
            "Getting a known AT from HashedId3 is not implemented yet!"
        )



import logging

from ..local_dynamic_map.ldm_constants import VAM
from .vam_ldm_adaptation import VRUBasicServiceLDM
from .vam_transmission_management import VAMTransmissionManagement, DeviceDataProvider
from ...btp.router import Router as BTPRouter
from .vam_coder import VAMCoder
from .vam_reception_management import VAMReceptionManagement
from ..local_dynamic_map.ldm_facility import LDMFacility


class VRUAwarenessService:
    """
    VRU Basis Service

    Attributes
    ----------
    btp_router : BTPRouter
        BTP Router.
    vam_coder : VAMCoder
        vam Coder.
    device_data_provider : DeviceDataProvider
        Vehicle Data.
    vam_transmission_management : vamTransmissionManagement
        vam Transmission Management.
    vam_reception_management : vamReceptionManagement
        vam Reception Management.
    """

    def __init__(
        self,
        btp_router: BTPRouter,
        device_data_provider: DeviceDataProvider,
        ldm: LDMFacility = None,
    ) -> None:
        """
        Initialize the Cooperative Awareness Basic Service.

        Parameters
        ----------
        btp_router : BTPRouter
            BTP Router.
        device_data_provider : DeviceDataProvider
            Vehicle Data.
        ldm: LDM Facility
            Local Dynamic Map Facility that will be used to provide data to the LDM.
        """
        self.logging = logging.getLogger("vru_basic_service")

        self.btp_router = btp_router
        self.vam_coder = VAMCoder()
        self.device_data_provider = device_data_provider
        vru_basic_service_ldm = None
        if ldm is not None:
            vru_basic_service_ldm = VRUBasicServiceLDM(ldm, [VAM], 5)

        self.vam_transmission_management = VAMTransmissionManagement(
            btp_router=btp_router,
            vam_coder=self.vam_coder,
            device_data_provider=self.device_data_provider,
            vru_basic_service_ldm=vru_basic_service_ldm,
        )
        self.vam_reception_management = VAMReceptionManagement(
            vam_coder=self.vam_coder,
            btp_router=self.btp_router,
            vru_basic_service_ldm=vru_basic_service_ldm,
        )

        self.logging.info("VRU Basic Service Started!")



class SendingException(Exception):
    """
    Exception raised when a packet cannot be sent.
    """


class PacketTooLongException(SendingException):
    """
    Exception raised when a packet is too long.
    """


class WindowsNotSupportedException(Exception):
    """
    Exception raised when the Windows OS is not supported.
    """


class InvalidMACAddressException(Exception):
    """
    Exception raised when the MAC address is invalid.
    """



"""
CA Reception Management.

This file contains the class for the CA Reception Management.
"""

import logging

from .cam_ldm_adaptation import CABasicServiceLDM
from .cam_coder import CAMCoder
from ...btp.service_access_point import BTPDataIndication
from ...btp.router import Router as BTPRouter


class CAMReceptionManagement:
    """
    This class is responsible for the CAM reception management.

    Attributes
    ----------
    cam_coder : CAMCoder
        CAM Coder object.
    btp_router : BTPRouter
        BTP Router object.
    ca_basic_service_ldm : CABasicServiceLDM
        CA Basic Service LDM.
    """

    def __init__(
        self,
        cam_coder: CAMCoder,
        btp_router: BTPRouter,
        ca_basic_service_ldm: CABasicServiceLDM = None,
    ) -> None:
        """
        Initialize the CAM Reception Management.

        Parameters
        ----------
        cam_coder : CAMCoder
            CAM Coder object.
        btp_router : BTPRouter
            BTP Router object.
        ldm: LDMFacility
            Local Dynamic Map where the data will be stashed.
        """
        self.logging = logging.getLogger("ca_basic_service")

        self.cam_coder = cam_coder
        self.btp_router = btp_router
        self.btp_router.register_indication_callback_btp(
            port=2001, callback=self.reception_callback
        )
        self.ca_basic_service_ldm = ca_basic_service_ldm

    def reception_callback(self, btp_indication: BTPDataIndication) -> None:
        """
        Callback for the reception of a CAM.

        Parameters
        ----------
        btp_indication : BTPDataIndication
            BTP Data Indication.
        """
        cam = self.cam_coder.decode(btp_indication.data)
        if self.ca_basic_service_ldm is not None:
            self.ca_basic_service_ldm.add_provider_data_to_ldm(cam)
        self.logging.debug(
            "Received CAM with timestamp: %s, station_id: %s",
            cam["cam"]["generationDeltaTime"],
            cam["header"]["stationID"],
        )



import logging
from .vam_coder import VAMCoder
from ...btp.service_access_point import BTPDataIndication
from ...btp.router import Router as BTPRouter

from .vam_ldm_adaptation import VRUBasicServiceLDM


class VAMReceptionManagement:
    """
    This class is responsible for the vam reception management.

    Attributes
    ----------
    vam_coder : vamCoder
        vam Coder object.

    """

    def __init__(
        self,
        vam_coder: VAMCoder,
        btp_router: BTPRouter,
        vru_basic_service_ldm: VRUBasicServiceLDM = None,
    ) -> None:
        """
        Initialize the vam Reception Management.

        Parameters
        ----------
        vam_coder : vamCoder
            vam Coder object.
        btp_router : BTPRouter
            BTP Router.
        ldm_facility : LDMFacility
            LDM Facility.
        """
        self.logging = logging.getLogger("vru_basic_service")
        self.vam_coder = vam_coder
        self.btp_router = btp_router
        self.btp_router.register_indication_callback_btp(
            port=2018, callback=self.reception_callback
        )
        self.vru_basic_service_ldm = vru_basic_service_ldm

    def reception_callback(self, btp_indication: BTPDataIndication) -> None:
        """
        Callback for the reception of a vam. Connected to LDM Facility in order to feed data.

        Parameters
        ----------
        btp_indication : BTPDataIndication
            BTP Data Indication.
        """
        vam = self.vam_coder.decode(btp_indication.data)
        if self.vru_basic_service_ldm is not None:
            self.vru_basic_service_ldm.add_provider_data_to_ldm(vam)
        self.logging.debug("Recieved message; %s", vam)
        self.logging.info(
            "Recieved VAM message with timestamp: %s, station_id: %s",
            vam["vam"]["generationDeltaTime"],
            vam["header"]["stationId"],
        )



"""
CAM Transmission Management

This file implements the CAM Transmission Management required by the CAM Basic Service.
"""

from math import trunc
import logging
from dateutil import parser
from .cam_coder import CAMCoder
from ...btp.router import Router as BTPRouter
from ...btp.service_access_point import (
    BTPDataRequest,
    CommonNH,
    PacketTransportType,
    CommunicationProfile,
    TrafficClass,
)
from .cam_ldm_adaptation import CABasicServiceLDM

T_GEN_CAM_MIN = 100  # T_GenCamMin [in ms]
T_GEN_CAM_MAX = 1000  # ms
T_CHECK_CAM_GEN = (
    T_GEN_CAM_MIN  # T_CheckCamGen [in ms] Shall be equal to or less than T_GenCamMin
)
T_GEN_CAM_DCC = (
    T_GEN_CAM_MIN  # T_GenCam_DCC [in ms] T_GenCamMin ≤ T_GenCam_DCC ≤ T_GenCamMax
)


class VehicleData:
    """
    Class that stores the vehicle data.

    Attributes
    ----------
    station_id : int
        Station Id as specified in ETSI TS 102 894-2 V2.3.1 (2024-08).
    station_type : int
        Station Type as specified in ETSI TS 102 894-2 V2.3.1 (2024-08).
    drive_direction : str
        Drive Direction as specified in ETSI TS 102 894-2 V2.3.1 (2024-08).
    vehicle_length : dict
        Vehicle Length as specified in ETSI TS 102 894-2 V2.3.1 (2024-08).
    vehicle_width : int
        Vehicle Width as specified in ETSI TS 102 894-2 V2.3.1 (2024-08).
    """

    def __init__(self) -> None:
        self.station_id = 0
        self.station_type = 0
        self.drive_direction = "unavailable"
        self.vehicle_length = {
            "vehicleLengthValue": 1023,
            "vehicleLengthConfidenceIndication": "unavailable",
        }
        self.vehicle_width = 62

    def set_station_type(self, station_type: int) -> None:
        """
        Set the station type.

        Parameters
        ----------
        station_type : int
            Station type.
        """
        if station_type < 0 or station_type > 15:
            raise ValueError("Station type must be between 0 and 15.")
        self.station_type = station_type

    def set_drive_direction(self, drive_direction: str) -> None:
        """
        Set the drive direction.

        Parameters
        ----------
        drive_direction : str
            Drive direction.
        """
        if drive_direction not in [
            "forward",
            "reverse",
            "unavailable",
        ]:
            raise ValueError(
                "Drive direction must be 'forward', 'reverse' or 'unavailable'."
            )
        self.drive_direction = drive_direction


class GenerationDeltaTime:
    """
    Generation Delta Time class. As specified in ETSI TS 102 894-2 V2.3.1 (2024-08).

    The reason this type is implemented as a class is to be able to quickly perform operations.

    Express the following way:
    generationDeltaTime = TimestampIts mod 65536
    TimestampIts represents an integer value in milliseconds since
    2004-01-01T00:00:00:000Z as defined in ETSI TS 102 894-2

    Attributes
    ----------
    msec : int
        Time in milliseconds.
    """

    def __init__(self) -> None:
        self.msec = 0

    def set_in_normal_timestamp(self, timestamp: float) -> None:
        """
        Set the Generation Delta Time in normal timestamp. [Seconds]

        Parameters
        ----------
        timestamp : int
            Timestamp in milliseconds.
        """
        self.msec = (timestamp - 1072911600000) % 65536

    def as_timestamp_in_certain_point(self, point_in_time: int) -> float:
        """
        Returns the generation delta time as timestamp as it would be if received at
        certain point in time.

        Parameters
        ----------
        point_in_time : int
            Timestamp in milliseconds

        Returns
        -------
        float
            Timestamp of the generation delta time in milliseconds
        """
        number_of_cycles = trunc((point_in_time - 1072911600000) / 65536)
        transformed_timestamp = self.msec + 65536 * number_of_cycles + 1072911600000
        if transformed_timestamp < point_in_time:
            return transformed_timestamp
        return self.msec + 65536 * (number_of_cycles - 1) + 1072911600000

    def __gt__(self, other: object) -> bool:
        """
        Greater than operator.
        """
        if isinstance(other, GenerationDeltaTime):
            return self.msec > other.msec
        return False

    def __lt__(self, other: object) -> bool:
        """
        Less than operator.
        """
        if isinstance(other, GenerationDeltaTime):
            return self.msec < other.msec
        return False

    def __ge__(self, other: object) -> bool:
        """
        Greater than or equal operator.
        """
        if isinstance(other, GenerationDeltaTime):
            return self.msec >= other.msec
        return False

    def __le__(self, other: object) -> bool:
        """
        Less than or equal operator.
        """
        if isinstance(other, GenerationDeltaTime):
            return self.msec <= other.msec
        return False

    def __add__(self, other: object) -> object:
        """
        Addition operator.
        """
        if isinstance(other, GenerationDeltaTime):
            return (self.msec + other.msec) % 65536
        return NotImplemented

    def __sub__(self, other: object) -> object:
        """
        Subtraction operator.
        """
        if isinstance(other, GenerationDeltaTime):
            subs = self.msec - other.msec
            if subs < 0:
                subs = subs + 65536
            return subs
        return NotImplemented


class CooperativeAwarenessMessage:
    """
    Cooperative Awareness Message class.

    Attributes
    ----------
    cam : dict
        All the CAM message in dict format as decoded by the CAMCoder.

    """

    def __init__(self) -> None:
        self.cam = self.generate_white_cam()

    def generate_white_cam(self) -> dict:
        """
        Generate a white CAM.
        """
        return {
            "header": {"protocolVersion": 2, "messageId": 2, "stationId": 0},
            "cam": {
                "generationDeltaTime": 0,
                "camParameters": {
                    "basicContainer": {
                        "stationType": 0,
                        "referencePosition": {
                            "latitude": 900000001,
                            "longitude": 1800000001,
                            "positionConfidenceEllipse": {
                                "semiMajorAxisLength": 4095,
                                "semiMinorAxisLength": 4095,
                                "semiMajorAxisOrientation": 3601,
                            },
                            "altitude": {
                                "altitudeValue": 800001,
                                "altitudeConfidence": "unavailable",
                            },
                        },
                    },
                    "highFrequencyContainer": (
                        "basicVehicleContainerHighFrequency",
                        {
                            "heading": {"headingValue": 3601, "headingConfidence": 127},
                            "speed": {"speedValue": 16383, "speedConfidence": 127},
                            "driveDirection": "unavailable",
                            "vehicleLength": {
                                "vehicleLengthValue": 1023,
                                "vehicleLengthConfidenceIndication": "unavailable",
                            },
                            "vehicleWidth": 62,
                            "longitudinalAcceleration": {
                                "value": 161,
                                "confidence": 102,
                            },
                            "curvature": {
                                "curvatureValue": 1023,
                                "curvatureConfidence": "unavailable",
                            },
                            "curvatureCalculationMode": "unavailable",
                            "yawRate": {
                                "yawRateValue": 32767,
                                "yawRateConfidence": "unavailable",
                            },
                        },
                    ),
                },
            },
        }

    def fullfill_with_vehicle_data(self, vehicle_data: VehicleData) -> None:
        """
        Fullfill the CAM with vehicle data.

        Parameters
        ----------
        vehicle_data : VehicleData
            Vehicle data.
        """
        self.cam["header"]["stationId"] = vehicle_data.station_id
        self.cam["cam"]["camParameters"]["basicContainer"][
            "stationType"
        ] = vehicle_data.station_type
        self.cam["cam"]["camParameters"]["highFrequencyContainer"][1][
            "driveDirection"
        ] = vehicle_data.drive_direction
        self.cam["cam"]["camParameters"]["highFrequencyContainer"][1][
            "vehicleLength"
        ] = vehicle_data.vehicle_length
        self.cam["cam"]["camParameters"]["highFrequencyContainer"][1][
            "vehicleWidth"
        ] = vehicle_data.vehicle_width

    def fullfill_gen_delta_time_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the generation delta time with the GPSD TPV data.

        Parameters
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "time" in tpv:
            self.cam["cam"]["generationDeltaTime"] = int(
                (((parser.parse(tpv["time"]).timestamp()) - 1072911600) * 1000) % 65536
            )

    def fullfill_basic_container_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the basic container with the GPSD TPV data.

        Parameters
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "lat" in tpv.keys():
            self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                "latitude"
            ] = int(tpv["lat"] * 10000000)
        if "lon" in tpv.keys():
            self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                "longitude"
            ] = int(tpv["lon"] * 10000000)
        if "epx" in tpv.keys() and "epy" in tpv.keys():
            self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                "positionConfidenceEllipse"
            ] = self.create_position_confidence(tpv["epx"], tpv["epy"])
        if "altHAE" in tpv.keys():
            alt = int(tpv["altHAE"] * 100)
            if alt < -800000:
                self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = -100000
            elif alt > 613000:
                self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = 800000
            else:
                self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = int(tpv["altHAE"] * 100)
        if "epv" in tpv.keys():
            self.cam["cam"]["camParameters"]["basicContainer"]["referencePosition"][
                "altitude"
            ]["altitudeConfidence"] = self.create_altitude_confidence(tpv["epv"])

    def fullfill_high_frequency_container_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the high frequency container with the GPSD TPV data.

        Parameters
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "track" in tpv.keys():
            self.cam["cam"]["camParameters"]["highFrequencyContainer"][1]["heading"][
                "headingValue"
            ] = int(tpv["track"])
        if "epd" in tpv.keys():
            self.cam["cam"]["camParameters"]["highFrequencyContainer"][1]["heading"][
                "headingConfidence"
            ] = self.create_heading_confidence(tpv["epd"])
        if "speed" in tpv.keys():
            if int(tpv["speed"] * 100) > 16381:
                self.cam["cam"]["camParameters"]["highFrequencyContainer"][1]["speed"][
                    "speedValue"
                ] = 16382
            else:
                self.cam["cam"]["camParameters"]["highFrequencyContainer"][1]["speed"][
                    "speedValue"
                ] = int(tpv["speed"] * 100)

    def fullfill_with_tpv_data(self, tpv: dict) -> None:
        """
        Convert a TPV data to a CAM.

        Parameters
        ----------
        tpv : dict
            GPSD TPV data.
        """
        self.fullfill_gen_delta_time_with_tpv_data(tpv)
        self.fullfill_basic_container_with_tpv_data(tpv)
        self.fullfill_high_frequency_container_with_tpv_data(tpv)

    def create_position_confidence(self, epx: int, epy: int) -> dict:
        """
        Translates the epx and epy TPV values to the position confidence ellipse value.

        Parameters
        ----------
        epx : int
            TPV epx value.
        epy : int
            TPV epy value.

        Returns
        -------
        dict
            Position confidence ellipse value.
        """
        position_confidence_ellipse = {
            "semiMajorAxisLength": int(epx * 100),
            "semiMinorAxisLength": int(epy * 100),
            "semiMajorAxisOrientation": 0,
        }
        if epy >= epx:
            position_confidence_ellipse = {
                "semiMajorAxisLength": int(epy * 100),
                "semiMinorAxisLength": int(epx * 100),
                "semiMajorAxisOrientation": 0,
            }
        return position_confidence_ellipse

    # def create_altitude_confidence(self, epv: float) -> str:
    #     """
    #     Translates the epv TPV value to the altitude confidence value.

    #     Parameters
    #     ----------
    #     epv : float
    #         TPV epv value.

    #     Returns
    #     -------
    #     str
    #         Altitude confidence value.
    #     """
    #     altitude_confidence = "unavailable"
    #     if epv < 0.01:
    #         altitude_confidence = "alt-000-01"
    #     elif epv < 0.02:
    #         altitude_confidence = "alt-000-02"
    #     elif epv < 0.05:
    #         altitude_confidence = "alt-000-05"
    #     elif epv < 0.1:
    #         altitude_confidence = "alt-000-10"
    #     elif epv < 0.2:
    #         altitude_confidence = "alt-000-20"
    #     elif epv < 0.5:
    #         altitude_confidence = "alt-000-50"
    #     elif epv < 1:
    #         altitude_confidence = "alt-001-00"
    #     elif epv < 2:
    #         altitude_confidence = "alt-002-00"
    #     elif epv < 5:
    #         altitude_confidence = "alt-005-00"
    #     elif epv < 10:
    #         altitude_confidence = "alt-010-00"
    #     elif epv < 20:
    #         altitude_confidence = "alt-020-00"
    #     elif epv < 50:
    #         altitude_confidence = "alt-050-00"
    #     elif epv < 100:
    #         altitude_confidence = "alt-100-00"
    #     elif epv <= 200:
    #         altitude_confidence = "alt-200-00"
    #     elif epv > 200:
    #         altitude_confidence = "outOfRange"
    #     return altitude_confidence

    def create_altitude_confidence(self, epv: float) -> str:
        """Translates the epv TPV value to the altitude confidence value.

        Parameters
        ----------
        epv : float
            TPV epv value.

        Returns
        -------
        str
            Altitude confidence value.
        """
        altitude_confidence_map = {
            0.01: "alt-000-01",
            0.02: "alt-000-02",
            0.05: "alt-000-05",
            0.1: "alt-000-10",
            0.2: "alt-000-20",
            0.5: "alt-000-50",
            1: "alt-001-00",
            2: "alt-002-00",
            5: "alt-005-00",
            10: "alt-010-00",
            20: "alt-020-00",
            50: "alt-050-00",
            100: "alt-100-00",
            200: "alt-200-00",
            float("inf"): "outOfRange",
        }

        for key in sorted(altitude_confidence_map.keys()):
            if epv < key:
                return altitude_confidence_map[key]

        return "unavailable"

    def create_heading_confidence(self, epd: float) -> int:
        """
        Translates the epd TPV value to the heading confidence value.

        Parameters
        ----------
        epd : float
            TPV epd value.

        Returns
        -------
        int
            Heading confidence value.
        """
        heading_confidence = 126
        if epd <= 12.5:
            heading_confidence = int(epd * 10)
        return heading_confidence

    def __str__(self) -> str:
        return str(self.cam)


class CAMTransmissionManagement:
    """
    CAM Transmission Management class.
    This sub-function ahould implement the protocol operation of the originating ITS-S, as specified
    in ETSI TS 102 894-2 V2.3.1 (2024-08) clause C.2, including in particular:
    - Activation and termination of CAM transmission operation.
    - Determination of the CAM generation frequency.
    - Trigger the generation of CAM.

    By now, it implements the same algorithms but being reactive to when a new position is received.

    Attributes
    ----------
    btp_router : BTPRouter
        BTP Router.
    vehicle_data : VehicleData
        Vehicle Data.
    cam_coder : CAMCoder
        CAM Coder.
    ca_basic_service_ldm : CABasicServiceLDM
        CA Basic Service LDM.
    t_gen_cam : int
        Time between CAM generations.
    last_cam_sent : CooperativeAwarenessMessage
        Last CAM sent.
    current_cam_to_send : CooperativeAwarenessMessage
        Current CAM to send.

    """

    def __init__(
        self,
        btp_router: BTPRouter,
        cam_coder: CAMCoder,
        vehicle_data: VehicleData,
        ca_basic_service_ldm: CABasicServiceLDM = None,
    ) -> None:
        """
        Initialize the CAM Transmission Management.
        """
        self.logging = logging.getLogger("ca_basic_service")
        self.btp_router: BTPRouter = btp_router
        self.vehicle_data = vehicle_data
        self.cam_coder = cam_coder
        self.ca_basic_service_ldm = ca_basic_service_ldm
        # self.T_GenCam_DCC = T_GenCamMin We don't have a DCC yet.
        self.t_gen_cam = T_GEN_CAM_MIN
        self.last_cam_sent = None
        self.current_cam_to_send = CooperativeAwarenessMessage()
        self.current_cam_to_send.fullfill_with_vehicle_data(self.vehicle_data)

    def location_service_callback(self, tpv: dict) -> None:
        """
        Callback function for location service.

        The Cooperative Awareness Service gets triggered everytime the location service gets a
        new position.

        TODO: Once the DCC is implemented, all conditions should be checked before sending a CAM.
        1) The time elapsed since the last CAM generation is equal to or greater than T_GenCam_Dcc,
        as applicable, and one of the following ITS-S dynamics related conditions is given:
            - the absolute difference between the current heading of the originating ITS-S and the
            heading included in the CAM previously transmitted by the originating ITS-S exceeds 4°;
            - the distance between the current position of the originating ITS-S and the position
            included in the CAM previously transmitted by the originating ITS-S exceeds 4 m;
            - the absolute difference between the current speed of the originating ITS-S and the
            speed included in the CAM previously transmitted by the originating ITS-S exceeds
            0,5 m/s.
        2) The time elapsed since the last CAM generation is equal to or greater than T_GenCam and,
        in the case of ITS-G5, is also equal to or greater than T_GenCam_Dcc.
        If one of the above two conditions is satisfied, a CAM shall be generated immediately.

        Parameters
        ----------
        tpv : dict
            GPSD TP
        """
        self.current_cam_to_send.fullfill_with_tpv_data(tpv)

        if self.last_cam_sent is None:
            self.send_next_cam()
            return
        received_generation_delta_time = GenerationDeltaTime()
        last_sent_generation_delta_time = GenerationDeltaTime()
        received_generation_delta_time.set_in_normal_timestamp(
            parser.parse(tpv["time"]).timestamp()
        )
        last_sent_generation_delta_time.msec = self.last_cam_sent.cam["cam"][
            "generationDeltaTime"
        ]
        if (
            received_generation_delta_time - last_sent_generation_delta_time
            >= self.t_gen_cam
        ):
            self.send_next_cam()
        else:
            pass

    def send_next_cam(self) -> None:
        """
        Send the next CAM.
        """
        if self.ca_basic_service_ldm is not None:
            self.ca_basic_service_ldm.add_provider_data_to_ldm(
                self.current_cam_to_send.cam
            )
        request = BTPDataRequest()
        request.btp_type = CommonNH.BTP_B
        request.destination_port = 2001
        request.gn_packet_transport_type = PacketTransportType()
        request.communication_profile = CommunicationProfile.UNSPECIFIED
        request.traffic_class = TrafficClass()
        request.data = self.cam_coder.encode(self.current_cam_to_send.cam)
        request.length = len(request.data)

        self.btp_router.btp_data_request(request)
        self.logging.debug(
            "Sent CAM message with timestamp: %d, station_id: %d",
            self.current_cam_to_send.cam["cam"]["generationDeltaTime"],
            self.current_cam_to_send.cam["header"]["stationId"],
        )

        self.last_cam_sent = self.current_cam_to_send
        self.current_cam_to_send = CooperativeAwarenessMessage()
        self.current_cam_to_send.fullfill_with_vehicle_data(self.vehicle_data)



import logging
from .denm_coder import DENMCoder
from .denm_transmission_management import DENMTransmissionManagement, VehicleData
from ...btp.router import Router as BTPRouter
from .denm_reception_management import DENMReceptionManagement
from ..local_dynamic_map.ldm_facility import LDMFacility


class DecentralizedEnvironmentalNotificationService:
    """
    Decentralized Environmental Notification (DEN) Service.

    Attributes
    ----------
    btp_router : BTPRouter
        BTP Router.
    denm_coder : DENMCoder
        DENM Coder.
    vehicle_data : VehicleData
        Vehicle Data.
    ldm : LDMFacility
        Local Dynamic Map (LDM) Service.
    denm_transmission_management : DENMTransmissionManagement
        DENM Transmission Management.
    denm_reception_management : DENMReceptionManagement
        DENM Reception Management.
    """

    def __init__(
        self, btp_router: BTPRouter, vehicle_data: VehicleData, ldm: LDMFacility = None
    ) -> None:
        """
        Initialize the Decentralized Environmental Notification Service.

        Parameters
        ----------
        btp_router : BTPRouter
            BTP Router.
        vehicle_data : VehicleData
            Vehicle Data.
        ldm: LDMFacility
            Local Dynamic Map (LDM) Service.
        """

        self.logging = logging.getLogger("denm_service")

        self.btp_router = btp_router
        self.denm_coder = DENMCoder()
        self.vehicle_data = vehicle_data
        self.ldm = ldm
        self.denm_transmission_management = DENMTransmissionManagement(
            btp_router=btp_router,
            denm_coder=self.denm_coder,
            vehicle_data=self.vehicle_data,
        )
        self.denm_reception_management = DENMReceptionManagement(
            denm_coder=self.denm_coder, btp_router=self.btp_router, ldm=ldm
        )

        self.logging.info("DENM Service Started!")



from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING
from ...facilities.local_dynamic_map.ldm_classes import ReferencePosition, TimestampIts

if TYPE_CHECKING:
    from .emergency_vehicle_approaching_service import (
        EmergencyVehicleApproachingService,
    )


class RelevanceArea:
    """
    Class to define the relevance area of the DENM.

    Parameters
    ----------
    relevance_distance : int
        Relevance distance from the vehicle to a traffic hazard
        or to its future position.
    relevance_direction : int
        Relevance direction from the vehicle to a traffic hazard
        or to its future position.
    """

    def __init__(self, r_distance: int, r_direction: int) -> None:
        """
        Initialize the relevance area of the DENM.

        Parameters
        ----------
        r_distance : int
            Relevance distance from the vehicle to a traffic hazard
            or to its future position.
        r_direction : int
            Relevance direction from the vehicle to a traffic hazard
            or to its future position.
        """
        self.relevance_distance = r_distance
        self.relevance_direction = r_direction


class PriorityLevel(Enum):
    """
    Class to define the priority levels of the DENMs.

    Parameters
    ----------
    AWARENESS : int
    WARNING : int
    PRECRASH : int
    """

    AWARENESS = 2
    WARNING = 1
    PRECRASH = 0


class DENRequest:
    """
    Class for storing the data of the DENM.

    Parameters
    ----------
    denm_interval : int
        Interval between two consecutive DENM messages in ms.
    priority_level : int
        Priority level of the DENM message.
    relevance_distance : str
        Relevance distance where the DENMs need to be received. [0] less than 50m
                                                                [1] -> 100m // [2] -> 200m // [3] -> 500m
    relevance_traffic_direction : str
        Relevance direction where the DENMs need to be received. [0] -> all traffic directions
                                                                 [1] -> upstream traffic    [2] -> downstream traffic
                                                                 [3] -> opposite direction traffic
    DENMTermination : str
        Type of termination message: isCancellation(0) or isNegation (1)
    detection_time : int
        Time of the detection of the hazard.
    time_period : int
        Max duration of the hazard.
    quality : int
        Quality level of the provided information. Values from 0 to 7. 7 = highest quality.
    event_position : Dict
        Position of the hazard. Contains Latitude, Longitude, Elevation and confidence values.
    heading : int
        Heading of the vehicle. Values from 0 to 3601. 0 = North.
    confidence : int
        Confidence level on provided data. % values from 0 to 100. 101 = not available.
    traceID : int
        [Optional] Provides the Trace ID enabling the providion of several traces.
    waypoints : Dict
        [Optional] Indication of the path followed by the vehicles before detecting the hazard
    rhs_cause_code : int
        rhs cause code defining the emergency vehicle approaching.
    rhs_subcause_code : int
        [OPTIONAL] RHS subcause code of the emergency vehicle approaching.
    rhs_event_speed : int
        Speed of the emergency vehicle.
    rhs_vehicle_type : int
        Type of emergency vehicle approaching. From ISO 3833, "Road vehicle Type - Terms and definitions".
        passengerCar = 0  //  minibus = 12
    rhs_trace : int
        [OPTIONAL] Set of planned waypoints for the vehicle.
    rhs_relevance_area : RelevanceArea
        Relevance area for the emergency vehicle approaching use case.
        Defines the farthest future position of the emergency vehicle.
    """

    def __init__(self) -> None:
        """
        Initialize the specific data of the DENM.
        """

        self.denm_interval = 100
        self.priority_level = 1
        # Data elements values
        self.detection_time = 0
        self.time_period = 0
        self.quality = 7
        self.event_position = {}
        self.heading = 0  # or N/A
        self.confidence = 2  # or N/A
        # self.traceID = N/A
        # self.waypoints = N/A
        # Emergency Vehicle Approaching specific data elements
        self.relevance_distance: str = None
        self.relevance_traffic_direction: str = None
        self.rhs_cause_code: str = None
        self.rhs_subcause_code: int = None
        self.rhs_event_speed: int = None
        self.rhs_vehicle_type: int = None
        # Longitudinal Collision Risk Warning specific data elements
        self.lcrw_cause_code: str = None
        self.lcrw_subcause_code: int = None

    def with_emergency_vehicle_approaching(
        self, service: "EmergencyVehicleApproachingService"
    ) -> None:
        """
        Fulfills the DENM Request for Emergency Vehicle Approaching Service

        Parameters
        ----------
        service : EmergencyVehicleApproachingService
            Emergency Vehicle Approaching Service object
        """
        self.denm_interval = service.denm_interval
        self.priority_level = service.priority_level

        # Relevance area parameters
        self.relevance_distance = "lessThan200m"
        self.relevance_traffic_direction = "upstreamTraffic"
        # self.DENMTermination = "isCancellation"

        # Data elements values
        self.detection_time = service.detection_time
        self.time_period = service.denm_duration
        self.event_position = service.event_position

        # Specific use cases data elemenets
        self.rhs_cause_code = "emergencyVehicleApproaching95"
        self.rhs_subcause_code = 1  # [OPTIONAL]
        self.rhs_event_speed = 30  # 108 km/h
        self.rhs_vehicle_type = 0
        # self.rhs_relevance_area = RelevanceArea(4, 0)

    def with_collision_risk_warning(
        self, detection_time: TimestampIts, event_position: ReferencePosition
    ) -> None:
        """
        Fulfills the DENM Request for Longitudinal Collision Risk Warnings

        Parameters
        ----------
        detection_time : TimestampIts
            Timestamp of the detection of the hazard.
        event_position : ReferencePosition
            Position of the hazard.
        """
        self.priority_level = 1
        self.detection_time = detection_time.timestamp
        self.event_position = event_position.to_dict()
        self.lcrw_cause_code = "collisionRisk97"  # Collision risk
        self.lcrw_subcause_code = 4  # Collision risk involving VRU
